"""
Module containing models for Zabbix protocol.
"""

from typing import List, Any, Optional, Dict, Union
import abc
import json
from ast import literal_eval
from uuid import uuid4


class _Model(abc.ABC):
    __slots__: List[str] = []

    def __init__(self) -> None:
        pass

    def __repr__(self) -> str:
        return str(ModelEncoder().default(self))


class ModelEncoder(json.JSONEncoder):
    """
    Class for encoding to JSON models implemented herein.
    """
    def default(self, o: _Model) -> Dict[str, Any]:
        d = {k: getattr(o, k) for k in type(o).__slots__ if getattr(o, k, False)}
        return d


class ItemData(_Model):
    """
    Class model representing data to be sent to a trapper item.

    Parameters
    ----------
    :host:
        Hostname to which the item belongs.
    :key:
        Item key
    :value:
        Value to be sent.
    :clock:
        Timestamp at which value was collected.
    """
    __slots__ = ['host', 'key', 'value', 'clock']

    def __init__(self, host: str, key: str, value: Any, clock: Optional[int] = None) -> None:
        super().__init__()
        self.host = host
        self.key = key
        self.value = value
        self.clock = clock


SenderData = ItemData


class AgentData(ItemData):
    """
    Class model representing data to be sent to a Zabbix agent (active) item.

    Parameters
    ----------
    :host:
        Hostname to which the item belongs.
    :key:
        Item key
    :value:
        Value to be sent.
    :clock:
        Timestamp at which value was collected.
    :ns:
        Nanoseconds for clock
    :state:
        State of an item. Set to 1 for Unsupported.

    Attributes
    ----------
    :id:
        Unique id for item within one session.
    """
    __slots__ = ItemData.__slots__ + ['ns', 'id', 'state']

    def __init__(self, host: str, key: str, value: Any, clock: int, ns: int, state: Optional[int] = None) -> None:
        super().__init__(host, key, value, clock)
        self.ns = ns
        self.id = 0
        self.state = state


class _TrapperRequest(_Model, abc.ABC):
    __slots__ = ['request', 'data', 'host', 'clock', 'ns', 'session']
    __supported_requests = ["active checks", "agent data", "sender data"]

    def __init__(self, request: str, **kwargs) -> None:
        super().__init__()
        if request not in _TrapperRequest.__supported_requests:
            raise ValueError
        self.request = request
        self.host = kwargs.get('host')
        self.data = kwargs.get('data', [])
        if self.data:
            self._check_items_classes(self.data, kwargs.get('item_class'))
        elif not self.data:
            self.data = []
        self.clock = kwargs.get('clock')
        self.ns = kwargs.get('ns')
        self.session = kwargs.get('session')

    def _check_items_classes(self, items, item_class):
        if not all(self._check_item_class(i, item_class) for i in items):
            raise TypeError

    def _check_item_class(self, item, item_class):
        return isinstance(item, item_class)


class ActiveChecksRequest(_TrapperRequest):
    """
    Class implementing protocol for requesting active checks for host.

    Parameters
    ----------
    :host:
        Get active checks for specified host.
    """
    def __init__(self, host: str) -> None:
        super().__init__(request="active checks", host=host)


class SenderDataRequest(_TrapperRequest):
    """
    Class implementing protocol for sending data with sender protocol.

    Parameters
    ----------
    :data:
        List of SenderData objects.
    """
    __item_class = SenderData

    def __init__(self, data: Optional[List[SenderData]] = None) -> None:
        super().__init__(
            request="sender data",
            data=data,
            item_class=SenderDataRequest.__item_class
            )

    def add_item(self, item: SenderData) -> None:
        """
        Add data to request.

        Parameters
        ----------
        :item:
            Instance of SenderData.
        """
        if not self._check_item_class(item, SenderDataRequest.__item_class):
            raise TypeError
        self.data.append(item)


class AgentDataRequest(_TrapperRequest):
    """
    Class implementing protocol for sending data gathered by active checks.
    Each instance should be used as unique data session.

    Parameters
    ----------
    :data:
        List of AgentData objects.
    """
    __item_class = AgentData

    def __init__(self, data: Optional[List[AgentData]] = None) -> None:
        super().__init__(
            request="agent data",
            data=data,
            item_class=AgentDataRequest.__item_class,
            session=uuid4().hex
            )

        self._item_id = 1
        if self.data:
            for d in self.data:
                d.id = self._item_id
                self._item_id += 1

    def add_item(self, item: AgentData) -> None:
        """
        Add data to request and assign an id to it.

        Parameters
        ----------
        :item:
            Instance of AgentData.
        """
        if not self._check_item_class(item, AgentDataRequest.__item_class):
            raise TypeError
        item.id = self._item_id
        self.data.append(item)
        self._item_id += 1


class ActiveItem:
    """
    Zabbix active item configuration.
    """

    __slots__ = ['key', 'delay', 'lastlogsize', 'mtime']

    def __init__(self, key: str, delay: int, lastlogsize: int = 0, mtime: int = 0) -> None:
        self.key = key
        self.delay = delay
        self.lastlogsize = lastlogsize
        self.mtime = mtime


class ServerResponse:
    """
    Class representing server responses.
    """
    def __init__(self, response: str) -> None:
        self.response = None
        self.data: List[ActiveItem] = []
        self._info: Dict[str, Any] = {}
        self._parse_response(response)

    @property
    def info(self) -> Union[Dict[str, Any], None]:
        return self._info if self._info else None

    def _parse_data(self, data):
        while data:
            item = data.pop()
            self.data.append(
                ActiveItem(item['key'],
                           item['delay'],
                           item['lastlogsize'],
                           item['mtime'])
            )

    def _parse_info(self, info):
        if not info:
            return None
        parts = (part.split(':') for part in info.split(';'))
        self._info = {k.strip(): literal_eval(v.strip()) for k, v in parts}

    def _parse_response(self, response):
        if response:
            loaded = json.loads(response)
            self.response = loaded['response']
            self._parse_info(loaded.get('info', None))
            self._parse_data(loaded.get('data', []))



"""
Module containing handlers for Zabbix protocol.
"""

from typing import Optional
import abc
import socket
import struct
import logging

logger = logging.getLogger(__name__)


class _Dstream(abc.ABC):
    def __init__(self, target: str, port: int = 10051, source_address: Optional[str] = None) -> None:
        self._ip = target
        self._port = port
        self._source_address = source_address

    def _send(self, payload: bytes) -> str:
        data = b""
        parsed = ""
        s = None
        try:
            if self._source_address:
                s = socket.create_connection(
                    (self._ip, self._port),
                    source_address=(self._source_address, 0))
                logger.info(f"Opening connection to {self._ip}:{self._port} with source address {self._source_address}")
            else:
                s = socket.create_connection((self._ip, self._port))
                logger.info(f"Opening connection to {self._ip}:{self._port}")
            packed = self._prepare_payload(payload)
            s.sendall(packed)
            data = self._recv_info(s)
            parsed = self._parse_response(data)
        except socket.error:
            logger.exception(f"Cannot connect to host {self._ip}:{self._port}:")
        except struct.error:
            logger.exception(f"Recived response is corrupted:")
        finally:
            if s:
                logger.info(f"Closing connection to {self._ip}:{self._port}")
                s.close()
            return parsed

    def _parse_response(self, response: bytes) -> str:
        _, length = struct.unpack('<5sQ', response[:13])
        data = struct.unpack(
            '<{}s'.format(length),
            response[13:13+length]
            )[0]

        return data.decode('utf-8')

    def _prepare_payload(self, payload: bytes) -> bytes:
        payload_len = len(payload)
        packed = struct.pack(
            '<5sQ{}s'.format(payload_len),
            b'ZBXD\x01',
            payload_len,
            payload
            )
        logger.debug(f"Packed payload for {self._ip}:{self._port}. Payload length: {payload_len}. Length with headers: {len(packed)}")
        return packed

    def _recv_info(self, socket_: socket.socket, buff: int = 1024) -> bytes:
        data = b""
        buffer = socket_.recv(buff)
        logger.debug(f"Received {len(buffer)} from {self._ip}:{self._port}")
        while buffer:
            data += buffer
            buffer = socket_.recv(buff)
            logger.debug(f"Received {len(buffer)} from {self._ip}:{self._port}")
        logger.debug(f"Completed data retrieval from {self._ip}:{self._port}. Total length: {len(data)}")
        return data



"""
Python implementation of Zabbix get.
"""

from typing import List, Dict, Optional
from zappix.dstream import _Dstream
import logging

logger = logging.getLogger(__name__)


class Get(_Dstream):
    """
    Class implementing zabbix_get utility.

    Parameters
    ----------
    :host:
        IP address of target host.
    :port:
        Port on which the Zabbix Agent listens.
    :source_address:
        Source IP address.
    """

    def __init__(self, host: str, port: int = 10050, source_address: Optional[str] = None) -> None:
        super().__init__(host, port, source_address)

    def _pack_key(self, key: str) -> bytes:
        _key = f"{key}\n".encode('utf-8')
        logging.info(f"Getting {key} from {self._ip}:{self._port}")
        return _key

    def get_value(self, key: str) -> str:
        """
        Get value of a single item identified by key.

        Parameters
        ----------
        :key:
            String representing an item key.

        Returns
        -------
        string
            Value of item.
        """

        return self._send(
            self._pack_key(key)
            )

    def get_report(self, keys: List[str]) -> Dict[str, str]:
        """
        Get value of a item identified by keys provided in supplied iterable.
        Note that there are executed synchronously, so the performance might not be ideal.

        Parameters
        ----------
        :keys:
            Iterable containing string representing item keys.

        Return
        ------
        dict
            Dict containing keys with corresponding values.
        """

        report = {key: self._send(self._pack_key(key)) for key in keys}
        return report


if __name__ == '__main__':
    import argparse
    params = argparse.ArgumentParser()
    params.add_argument('-s', '--host', nargs='?')
    params.add_argument('-p', '--port', nargs='?', default=10050, type=int)
    params.add_argument('-I', '--source-address', nargs='?')
    params.add_argument('-k', '--key', nargs='?')
    args = params.parse_args()

    if args.source_address:
        zab = Get(args.host, args.port, args.source_address)
    else:
        zab = Get(args.host, args.port)

    result = zab.get_value(args.key)
    print(result)



from typing import List, Optional
from zappix.protocol import ActiveChecksRequest, AgentDataRequest, ModelEncoder, ActiveItem
from zappix.protocol import ServerResponse
from zappix.dstream import _Dstream
import json
import logging

logger = logging.getLogger(__name__)


class AgentActive(_Dstream):
    """
    Class for getting active check configuration for a host from Zabbix Server
    and sending collected data.

    Parameters
    ----------
    :host:
        Technical hostname as configured in Zabbix.
    :server:
        IP address of target Zabbix Server.
    :server_port:
        Port on which the Zabbix Server listens.
    :source_address:
        Source IP address.
    """

    def __init__(self, host: str, server: str, server_port: int = 10051, source_address: Optional[str] = None) -> None:
        super().__init__(server, server_port, source_address)
        self._host = host

    def get_active_checks(self) -> List[ActiveItem]:
        """
        Gets list of active checks for host.

        Returns
        -------
        list
            List of ActiveItem objects.
        """
        request = ActiveChecksRequest(self._host)
        logger.info(f"Getting active checks for host: {self._host} from: {self._ip}:{self._port}")
        result = self._send(
            json.dumps(request, cls=ModelEncoder).encode("utf-8")
            )
        return ServerResponse(result).data

    def send_collected_data(self, data: AgentDataRequest) -> ServerResponse:
        """
        Sends collected data to Zabbix.

        Parameters
        ----------
        :data:
            Instance of AgentDataRequest.

        Returns
        -------
        list
            List of ActiveItem objects.
        """
        if not isinstance(data, AgentDataRequest):
            logger.error(f"Object {data} is not an instance AgentDataRequest")
            raise ValueError
        result = self._send(
            json.dumps(data, cls=ModelEncoder).encode("utf-8")
            )
        return ServerResponse(result)



"""
Python implementation of Zabbix sender.
"""

from typing import List, Any, Optional, Dict, Tuple, Callable, Union
from zappix.dstream import _Dstream
from zappix.protocol import (SenderData,
                             SenderDataRequest,
                             ModelEncoder,
                             ServerResponse)
import json
import csv
import time
import functools
import logging

logger = logging.getLogger(__name__)


class Sender(_Dstream):
    """
    Class implementing zabbix_sender utility.

    Parameters
    ----------
    :server:
        IP address of target Zabbix Server.
    :port:
        Port on which the Zabbix Server listens.
    :source_address:
        Source IP address.
    """

    def __init__(self, server: str, port: int = 10051, source_address: Optional[str] = None) -> None:
        super().__init__(server, port, source_address)

    def send_value(self, host: str, key: str, value: Any) -> Union[Dict[str, Any], None]:
        """
        Send a single value to a Zabbix host.

        Parameters
        ----------
        :host:
            Name of a host as visible in Zabbix frontend.
        :key:
            String representing an item key.
        :value:
            Value to be sent.

        Returns
        -------
        dict
            Information from server.
        """
        payload = SenderDataRequest()
        payload.add_item(SenderData(host, key, value))

        response = self._send(json.dumps(payload, cls=ModelEncoder).encode("utf-8"))
        return ServerResponse(response).info

    def send_file(self, file: str, with_timestamps: bool = False) -> Tuple[Union[Dict[str, Any], None], List[int]]:
        """
        Send values contained in a file to specified hosts.

        Parameters
        ----------
        :file:
            Path to file with data.
        :with_timestamps:
            Specify whether file contains timestamps for items.

        Returns
        -------
        dict
            Information from server.
        """
        payload, corrupted_lines = self._parse_file(file, with_timestamps)
        response = self._send(json.dumps(payload, cls=ModelEncoder).encode("utf-8"))
        return ServerResponse(response).info, corrupted_lines

    def send_result(self, host: str, key: str) -> Any:
        """
        Decorator that sends a functions resoult to specified key on a host.

        Parameters
        ----------
        :host:
            Name of a host as visible in Zabbix frontend.
        :key:
            String representing an item key.

        Returns
        -------
        dynamic
            Result of decorated function.
        """
        def wrap_function(func: Callable) -> Callable:
            @functools.wraps(func)
            def get_value(*args, **kwargs):
                result = func(*args, **kwargs)
                self.send_value(host, key, str(result))
                return result
            return get_value
        return wrap_function

    def _parse_file(self, file: str, with_timestamps: bool = False) -> Tuple[SenderDataRequest, List[int]]:
        with open(file, 'r', encoding='utf-8') as values:
            payload = SenderDataRequest()
            reader = csv.reader(values, delimiter=' ', skipinitialspace=True)
            logger.info(f"Reading data from {file}")
            failed_lines = []

            for row in reader:
                try:
                    if with_timestamps:
                        data = SenderData(row[0], row[1], row[3], int(row[2]))
                    else:
                        data = SenderData(row[0], row[1], row[2])
                except (IndexError, ValueError):
                    failed_lines.append(reader.line_num)
                    logger.exception(f"Could not parse {file} at line {reader.line_num}")
                else:
                    if all(row):
                        payload.add_item(data)
                        logger.debug(f"Adding {data} to Sender payload")
                    else:
                        failed_lines.append(reader.line_num)
        if with_timestamps:
            now = time.time()
            payload.clock = int(now//1)
            payload.ns = int(now % 1 * 1e9)

        return payload, failed_lines


if __name__ == '__main__':
    import argparse
    params = argparse.ArgumentParser()
    params.add_argument('-z', '--zabbix', nargs='?')
    params.add_argument('-p', '--port', nargs='?', default=10051, type=int)
    params.add_argument('-I', '--source-address', nargs='?')
    params.add_argument('-s', '--host', nargs='?')
    params.add_argument('-k', '--key', nargs='?')
    params.add_argument('-o', '--value', nargs='?')
    params.add_argument('-i', '--input-file', nargs='?')
    params.add_argument('-T', '--with-timestamps', action='store_true')
    args = params.parse_args()

    if args.source_address:
        zab = Sender(args.zabbix, args.port, args.source_address)
    else:
        zab = Sender(args.zabbix, args.port)

    if all([args.host, args.key, args.value]):
        result = zab.send_value(args.host, args.key, args.value)
    elif args.input_file:
        result, corrupted_lines = zab.send_file(args.input_file, True if args.with_timestamps else False)

    print('info from server: "{}"'.format(result))



"""
Project home:
https://gitlab.com/szuro/zappix
"""

from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), 'r', encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='zappix',
    version='0.3.3',
    description='A Python replacement for Zabbix sender and get.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://gitlab.com/szuro/zappix',
    author='Robert Szulist',
    author_email='r.szulist@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: System :: Monitoring',
        'License :: OSI Approved :: MIT License',
    ],
    keywords='zabbix get sender',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    test_suite="tests",
    extras_require={
        'dev': [
            'tox',
            'pyzabbix'
        ]
    }
)
